@@ -43,36 +43,48 @@ extern "C" {
 
 namespace libdnf {
 
-static void setSovable(Pool * pool, Solvable *solvable, std::string name,
-    std::string stream, std::string version, std::string context, const char * arch)
+/**
+ * @brief create solvable with:
+ *   Name: $name:$stream:$context
+ *   Version: $version
+ *   Arch: $arch (If arch is not defined, set "noarch")
+ *   Provides: module($name)
+ *   Provides: module($name:$stream)
+ *   Summary: original_context
+ *   Description: name:stream
+ */
+static void setSovable(Pool * pool, Solvable * solvable, const std::string & name,
+    const std::string & stream, const std::string & version, const std::string & context, const char * arch, const std::string & original_context)
 {
     std::ostringstream ss;
-    // create solvable with:
-    //   Name: $name:$stream:$version:$context
-    //   Version: 0
-    //   Arch: $arch
-    ss << name << ":" << stream << ":" << version << ":" << context;
+    //   Name: $name:$stream:$context
+    ss << name << ":" << stream << ":" << context;
     solvable_set_str(solvable, SOLVABLE_NAME, ss.str().c_str());
-    solvable_set_str(solvable, SOLVABLE_EVR, "0");
+    solvable_set_str(solvable, SOLVABLE_EVR, version.c_str());
     // TODO Test can be remove when modules will be always with arch
     solvable_set_str(solvable, SOLVABLE_ARCH, arch ? arch : "noarch");
+
+    // store original context in summary
+    solvable_set_str(solvable, SOLVABLE_SUMMARY, original_context.c_str());
+
+    // store original name:stream in description
+    ss.str(std::string());
+    ss << name << ":" << stream;
+    solvable_set_str(solvable, SOLVABLE_DESCRIPTION, ss.str().c_str());
+
     // create Provide: module($name)
     ss.str(std::string());
     ss << "module(" << name << ")";
     auto depId = pool_str2id(pool, ss.str().c_str(), 1);
     solvable_add_deparray(solvable, SOLVABLE_PROVIDES, depId, -1);
+    // create Conflicts: module($name)
+    solvable_add_deparray(solvable, SOLVABLE_CONFLICTS, depId, 0);
 
     // create Provide: module($name:$stream)
     ss.str(std::string());
     ss << "module(" << name << ":" << stream << ")";
     depId = pool_str2id(pool, ss.str().c_str(), 1);
     solvable_add_deparray(solvable, SOLVABLE_PROVIDES, depId, -1);
-
-    // create Provide: module($name:$stream:$version)
-    ss.str(std::string());
-    ss << "module(" << name << ":" << stream << ":" << version << ")";
-    depId = pool_str2id(pool, ss.str().c_str(), 1);
-    solvable_add_deparray(solvable, SOLVABLE_PROVIDES, depId, -1);
 }
 
 
@@ -86,7 +98,7 @@ static std::pair<std::string, std::string> parsePlatform(const std::string & pla
 }
 
 ModulePackage::ModulePackage(DnfSack * moduleSack, LibsolvRepo * repo,
-    ModulemdModuleStream * mdStream, const std::string & repoID)
+    ModulemdModuleStream * mdStream, const std::string & repoID, const std::string & context)
         : mdStream(mdStream)
         , moduleSack(moduleSack)
         , repoID(repoID)
@@ -97,8 +109,9 @@ ModulePackage::ModulePackage(DnfSack * moduleSack, LibsolvRepo * repo,
     Pool * pool = dnf_sack_get_pool(moduleSack);
     id = repo_add_solvable(repo);
     Solvable *solvable = pool_id2solvable(pool, id);
-
-    setSovable(pool, solvable, getName(), getStream(), getVersion(), getContext(), getArchCStr());
+    std::string original_context = getContext();
+    setSovable(pool, solvable, getName(), getStream(), getVersion(), context.empty() ? original_context : context,
+               getArchCStr(), original_context);
     createDependencies(solvable);
     HyRepo hyRepo = static_cast<HyRepo>(repo->appdata);
     libdnf::repoGetImpl(hyRepo)->needs_internalizing = 1;
@@ -150,8 +163,8 @@ void ModulePackage::createDependencies(Solvable *solvable) const
     Pool * pool = dnf_sack_get_pool(moduleSack);
 
     for (const auto &dependency : getModuleDependencies()) {
-        for (const auto &requires : dependency.getRequires()) {
-            for (const auto &singleRequires : requires) {
+        for (const auto &dep_requires : dependency.getRequires()) {
+            for (const auto &singleRequires : dep_requires) {
                 auto moduleName = singleRequires.first;
                 std::vector<std::string> requiresStream;
                 for (const auto &moduleStream : singleRequires.second) {
@@ -179,7 +192,7 @@ void ModulePackage::createDependencies(Solvable *solvable) const
                     ss << "(";
                     ss << std::accumulate(std::next(requiresStream.begin()),
                                             requiresStream.end(), requiresStream[0],
-                                            [](std::string & a, std::string & b)
+                                            [](const std::string & a, const std::string & b)
                                             { return a + " or " + b; });
                     ss << ")";
                     depId = pool_parserpmrichdep(pool, ss.str().c_str());
@@ -192,11 +205,64 @@ void ModulePackage::createDependencies(Solvable *solvable) const
     }
 }
 
+std::vector<std::string> ModulePackage::getRequires(ModulemdModuleStream * mdStream, bool removePlatform)
+{
+    std::vector<std::string> dependencies_result;
+
+    GPtrArray * cDependencies = modulemd_module_stream_v2_get_dependencies((ModulemdModuleStreamV2 *) mdStream);
+
+    for (unsigned int i = 0; i < cDependencies->len; i++) {
+        auto dependencies = static_cast<ModulemdDependencies *>(g_ptr_array_index(cDependencies, i));
+        if (!dependencies) {
+            continue;
+        }
+        char** runtimeReqModules = modulemd_dependencies_get_runtime_modules_as_strv(dependencies);
+
+        for (char **iterModule = runtimeReqModules; iterModule && *iterModule; iterModule++) {
+            char ** runtimeReqStreams = modulemd_dependencies_get_runtime_streams_as_strv(dependencies, *iterModule);
+            auto moduleName = static_cast<char *>(*iterModule);
+            if (removePlatform && strcmp(moduleName, "platform") == 0) {
+                g_strfreev(runtimeReqStreams);
+                continue;
+            }
+            std::ostringstream ss;
+            std::vector<std::string> requiredStreams;
+            for (char **iterStream = runtimeReqStreams; iterStream && *iterStream; iterStream++) {
+                requiredStreams.push_back(*iterStream);
+            }
+            if (requiredStreams.empty()) {
+                dependencies_result.emplace_back(moduleName);
+            } else {
+                std::sort(requiredStreams.begin(), requiredStreams.end());
+                ss << moduleName << ":" << "[" << *requiredStreams.begin();
+                for (auto iter = std::next(requiredStreams.begin()); iter != requiredStreams.end(); ++iter) {
+                    ss << "," << *iter;
+                }
+                ss << "]";
+                dependencies_result.emplace_back(std::move(ss.str()));
+            }
+            g_strfreev(runtimeReqStreams);
+        }
+        g_strfreev(runtimeReqModules);
+    }
+
+    return dependencies_result;
+}
+
 std::string ModulePackage::getYaml() const
 {
     ModulemdModuleIndex * i = modulemd_module_index_new();
     modulemd_module_index_add_module_stream(i, mdStream, NULL);
-    return modulemd_module_index_dump_to_string(i, NULL);
+    gchar *cStrYaml = modulemd_module_index_dump_to_string(i, NULL);
+    std::string yaml = std::string(cStrYaml);
+    g_free(cStrYaml);
+    g_object_unref(i);
+    return yaml;
+}
+
+bool ModulePackage::getStaticContext() const
+{
+    return modulemd_module_stream_v2_is_static_context((ModulemdModuleStreamV2 *) mdStream);
 }
 
 /**
@@ -215,18 +281,6 @@ std::string ModulePackage::getName() const
     return name ? name : "";
 }
 
-/**
- * @brief Return module $name:$stream.
- *
- * @return std::string
- */
-std::string ModulePackage::getNameStream() const
-{
-    std::ostringstream ss;
-    ss << getName() << ":" << getStream();
-    return ss.str();
-}
-
 /**
  * @brief Return module $name:$stream:$version.
  *
@@ -364,6 +418,24 @@ std::vector<std::string> ModulePackage::getArtifacts() const
     return result_rpms;
 }
 
+/**
+ * @brief Return sorted list of RPM names that are demodularized.
+ *
+ * @return std::vector<std::string>
+ */
+std::vector<std::string> ModulePackage::getDemodularizedRpms() const
+{
+    std::vector<std::string> result_rpms;
+    char ** rpms = modulemd_module_stream_v2_get_demodularized_rpms((ModulemdModuleStreamV2 *) mdStream);
+
+    for (char **iter = rpms; iter && *iter; iter++) {
+        result_rpms.emplace_back(std::string(*iter));
+    }
+
+    g_strfreev(rpms);
+    return result_rpms;
+}
+
 std::vector<ModuleProfile>
 ModulePackage::getProfiles(const std::string &name) const
 {
@@ -388,6 +460,39 @@ ModulePackage::getProfiles(const std::string &name) const
     return result_profiles;
 }
 
+/* @brief Return default profiles as defined in the modulemd itself.
+ *
+ * Note this is probably not the function you want. You likely want to use
+ * ModulePackageContainer::getDefaultProfiles() instead, which sources the
+ * distro-level modulemd-defaults.
+ *
+ * Also, this function returns the first default profile, but instead we want it
+ * to return all default profiles. So supporting this properly in the future
+ * would require a new ModulePackage::getDefaultProfiles() which returns an
+ * std::vec<ModuleProfile> instead.
+ *
+ * @return ModuleProfile
+ */
+ModuleProfile
+ModulePackage::getDefaultProfile() const
+{
+    //TODO(amatej): replace with
+    //char ** profiles = modulemd_module_stream_v2_search_profiles((ModulemdModuleStreamV2 *) mdStream, profileNameCStr);
+    char ** profiles = modulemd_module_stream_v2_get_profile_names_as_strv((ModulemdModuleStreamV2 *) mdStream);
+    if (g_strv_length (profiles) == 1) {
+        return ModuleProfile(modulemd_module_stream_v2_get_profile((ModulemdModuleStreamV2 *) mdStream, profiles[0]));
+    }
+
+    for (char **iter = profiles; iter && *iter; iter++) {
+        auto profile = ModuleProfile(modulemd_module_stream_v2_get_profile((ModulemdModuleStreamV2 *) mdStream, *iter));
+        if (profile.isDefault()) {
+            return profile;
+        }
+    }
+
+    throw std::runtime_error("No default profile found for " + getFullIdentifier());
+}
+
 /**
  * @brief Return list of ModuleProfiles.
  *
@@ -579,7 +684,7 @@ ModulePackage::createPlatformSolvable(DnfSack * sack, DnfSack * moduleSack,
     repoImpl->needs_internalizing = 1;
     Id id = repo_add_solvable(repo);
     Solvable *solvable = pool_id2solvable(pool, id);
-    setSovable(pool, solvable, name, stream, version, context, "noarch");
+    setSovable(pool, solvable, name, stream, version, context, "noarch", context);
     repoImpl->needs_internalizing = 1;
     dnf_sack_set_provides_not_ready(moduleSack);
     dnf_sack_set_considered_to_update(moduleSack);
@@ -587,4 +692,13 @@ ModulePackage::createPlatformSolvable(DnfSack * sack, DnfSack * moduleSack,
     return id;
 }
 
+std::string ModulePackage::getNameStream(ModulemdModuleStream * mdStream)
+{
+    std::ostringstream ss;
+    auto name = modulemd_module_stream_get_module_name(mdStream);
+    auto stream = modulemd_module_stream_get_stream_name(mdStream);
+    ss << (name ? name : "") << ":" << (stream ? stream : "");
+    return ss.str();
+}
+
 }
