@@ -49,6 +49,45 @@ extern "C" {
 #include "modulemd/ModuleMetadata.hpp"
 #include "modulemd/ModuleProfile.hpp"
 
+
+namespace {
+
+/// Requires resolved goal
+/// Takes listInstalls() from goal and keep solvables with the solvable-name (<name>:<stream>:<context>) in query 
+void goal2name_query(libdnf::Goal & goal, libdnf::Query & query)
+{
+    auto pool = dnf_sack_get_pool(goal.getSack());
+    auto installList = goal.listInstalls();
+    std::vector<const char *> module_names;
+    Id id = -1;
+    while ((id = installList.next(id)) != -1) {
+        Solvable * s = pool_id2solvable(pool, id);
+        const char * name = pool_id2str(pool, s->name);
+        module_names.push_back(name);
+    }
+    module_names.push_back(nullptr);
+    query.addFilter(HY_PKG_NAME, HY_EQ, module_names.data());
+}
+
+/**
+ * @brief In python => ";".join(list.sort())
+ */
+std::string concentrateVectorString(std::vector<std::string> & list)
+{
+    if (list.empty()) {
+        return {};
+    }
+    std::sort(list.begin(), list.end());
+    std::ostringstream ss;
+    ss << *list.begin();
+    for (auto require = std::next(list.begin()); require != list.end(); ++require) {
+        ss << ";" << *require;
+    }
+    return ss.str();
+}
+
+}
+
 namespace std {
 
 template<>
@@ -150,12 +189,22 @@ public:
         const std::vector<ModulePackage *> & modules, bool debugSolver);
     bool insert(const std::string &moduleName, const char *path);
     std::vector<ModulePackage *> getLatestActiveEnabledModules();
+    /// Required to call after all modules v3 are in metadata
+    void addVersion2Modules();
 
 private:
     friend struct ModulePackageContainer;
     class ModulePersistor;
     std::unique_ptr<ModulePersistor> persistor;
     std::map<Id, std::unique_ptr<ModulePackage>> modules;
+    /// Internal sack with module solvables
+    /// resolveContext = <moduleContext> if moduleMdVersion > 2, else generated from requires
+    /// solvable.name = <moduleName>:<moduleStream>:<resolveContext>
+    /// solvable.evr = <moduleVersion>
+    /// solvable.arch = <moduleArch>
+    /// solvable.summary = <moduleContext>
+    /// solvable.description = <moduleName>:<moduleStream>
+    /// solvable.conflicts = module(<moduleName>)
     DnfSack * moduleSack;
     std::unique_ptr<PackageSet> activatedModules;
     std::string installRoot;
@@ -163,6 +212,8 @@ private:
     ModuleMetadata moduleMetadata;
 
     std::map<std::string, std::string> moduleDefaults;
+    std::vector<std::tuple<LibsolvRepo *, ModulemdModuleStream *, std::string>> modulesV2;
+
     bool isEnabled(const std::string &name, const std::string &stream);
 };
 
@@ -205,6 +256,7 @@ private:
         std::vector<std::string> profiles;
         ModuleState state;
         bool locked;
+        int streamChangesNum;
     };
     std::pair<ConfigParser, struct Config> & getEntry(const std::string & moduleName);
     bool update(const std::string &name);
@@ -234,13 +286,6 @@ ModulePackageContainer::ModulePackageContainer(bool allArch, std::string install
         pImpl->persistDir = dir;
     }
 
-    Pool * pool = dnf_sack_get_pool(pImpl->moduleSack);
-    HyRepo hrepo = hy_repo_create("available");
-    auto repoImpl = libdnf::repoGetImpl(hrepo);
-    LibsolvRepo *repo = repo_create(pool, "available");
-    repo->appdata = hrepo;
-    repoImpl->libsolvRepo = repo;
-    repoImpl->needs_internalizing = 1;
     pImpl->installRoot = installRoot;
     g_autofree gchar * path = g_build_filename(pImpl->installRoot.c_str(),
                                               "/etc/dnf/modules.d", NULL);
@@ -323,23 +368,36 @@ ModulePackageContainer::add(const std::string &fileContent, const std::string &
     md.addMetadataFromString(fileContent, 0);
     md.resolveAddedMetadata();
 
+    LibsolvRepo * repo = nullptr;
     LibsolvRepo * r;
     Id id;
 
+    // Search whether available repo was already created
     FOR_REPOS(id, r) {
-        if (strcmp(r->name, "available") == 0) {
-            g_autofree gchar * path = g_build_filename(pImpl->installRoot.c_str(),
-                                                      "/etc/dnf/modules.d", NULL);
-            std::vector<ModulePackage *> packages = md.getAllModulePackages(pImpl->moduleSack, r, repoID);
-            for(auto const& modulePackagePtr: packages) {
-                std::unique_ptr<ModulePackage> modulePackage(modulePackagePtr);
-                pImpl->modules.insert(std::make_pair(modulePackage->getId(), std::move(modulePackage)));
-                pImpl->persistor->insert(modulePackagePtr->getName(), path);
-            }
-
-            return;
+        if (strcmp(r->name, repoID.c_str()) == 0) {
+            repo = r;
         }
     }
+
+    // If not created yet, create it
+    if (!repo) {
+        Pool * pool = dnf_sack_get_pool(pImpl->moduleSack);
+        HyRepo hrepo = hy_repo_create(repoID.c_str());
+        auto repoImpl = libdnf::repoGetImpl(hrepo);
+        repo = repo_create(pool, repoID.c_str());
+        repo->appdata = hrepo;
+        repoImpl->libsolvRepo = repo;
+        repoImpl->needs_internalizing = 1;
+    }
+
+    // add all modules to repository and pass ownership to module container
+    g_autofree gchar * path = g_build_filename(pImpl->installRoot.c_str(), "/etc/dnf/modules.d", NULL);
+    auto packages = md.getAllModulePackages(pImpl->moduleSack, repo, repoID, pImpl->modulesV2);
+    for(auto const& modulePackagePtr: packages) {
+        std::unique_ptr<ModulePackage> modulePackage(modulePackagePtr);
+        pImpl->modules.insert(std::make_pair(modulePackage->getId(), std::move(modulePackage)));
+        pImpl->persistor->insert(modulePackagePtr->getName(), path);
+    }
 }
 
 Id
@@ -376,6 +434,7 @@ void ModulePackageContainer::createConflictsBetweenStreams()
 
 bool ModulePackageContainer::empty() const noexcept
 {
+    pImpl->addVersion2Modules();
     return pImpl->modules.empty();
 }
 
@@ -469,11 +528,13 @@ bool ModulePackageContainer::isDisabled(const ModulePackage * module)
 std::vector<std::string> ModulePackageContainer::getDefaultProfiles(std::string moduleName,
     std::string moduleStream)
 {
+    pImpl->addVersion2Modules();
     return pImpl->moduleMetadata.getDefaultProfiles(moduleName, moduleStream);
 }
 
 const std::string & ModulePackageContainer::getDefaultStream(const std::string &name) const
 {
+    pImpl->addVersion2Modules();
     auto it = pImpl->moduleDefaults.find(name);
     if (it == pImpl->moduleDefaults.end()) {
         return EMPTY_RESULT;
@@ -483,6 +544,7 @@ const std::string & ModulePackageContainer::getDefaultStream(const std::string &
 
 const std::string & ModulePackageContainer::getEnabledStream(const std::string &name)
 {
+    pImpl->addVersion2Modules();
     return pImpl->persistor->getStream(name);
 }
 
@@ -490,8 +552,12 @@ const std::string & ModulePackageContainer::getEnabledStream(const std::string &
  * @brief Mark ModulePackage as part of an enabled stream.
  */
 bool
-ModulePackageContainer::enable(const std::string &name, const std::string & stream)
+ModulePackageContainer::enable(const std::string &name, const std::string & stream, const bool count)
 {
+    pImpl->addVersion2Modules();
+    if (count) {
+        pImpl->persistor->getEntry(name).second.streamChangesNum++;
+    }
     bool changed = pImpl->persistor->changeStream(name, stream);
     if (pImpl->persistor->changeState(name, ModuleState::ENABLED)) {
         changed = true;
@@ -504,41 +570,50 @@ ModulePackageContainer::enable(const std::string &name, const std::string & stre
 }
 
 bool
-ModulePackageContainer::enable(const ModulePackage * module)
+ModulePackageContainer::enable(const ModulePackage * module, const bool count)
 {
-    return enable(module->getName(), module->getStream());
+    return enable(module->getName(), module->getStream(), count);
 }
 
 /**
  * @brief Mark module as not part of an enabled stream.
  */
-void ModulePackageContainer::disable(const std::string & name)
+void ModulePackageContainer::disable(const std::string & name, const bool count)
 {
+    pImpl->addVersion2Modules();
+    if (count) {
+        pImpl->persistor->getEntry(name).second.streamChangesNum++;
+    }
+
     pImpl->persistor->changeState(name, ModuleState::DISABLED);
     pImpl->persistor->changeStream(name, "");
     auto & profiles = pImpl->persistor->getEntry(name).second.profiles;
     profiles.clear();
 }
 
-void ModulePackageContainer::disable(const ModulePackage * module)
+void ModulePackageContainer::disable(const ModulePackage * module, const bool count)
 {
-    disable(module->getName());
+    disable(module->getName(), count);
 }
 
 /**
  * @brief Reset module state so it's no longer enabled or disabled.
  */
-void ModulePackageContainer::reset(const std::string & name)
+void ModulePackageContainer::reset(const std::string & name, const bool count)
 {
+    pImpl->addVersion2Modules();
+    if (count) {
+        pImpl->persistor->getEntry(name).second.streamChangesNum++;
+    }
     pImpl->persistor->changeState(name, ModuleState::UNKNOWN);
     pImpl->persistor->changeStream(name, "");
     auto & profiles = pImpl->persistor->getEntry(name).second.profiles;
     profiles.clear();
 }
 
-void ModulePackageContainer::reset(const ModulePackage * module)
+void ModulePackageContainer::reset(const ModulePackage * module, const bool count)
 {
-    reset(module->getName());
+    reset(module->getName(), count);
 }
 
 /**
@@ -570,6 +645,7 @@ bool ModulePackageContainer::isChanged()
 void ModulePackageContainer::install(const std::string &name, const std::string &stream,
     const std::string &profile)
 {
+    pImpl->addVersion2Modules();
     for (const auto &iter : pImpl->modules) {
         auto modulePackage = iter.second.get();
         if (modulePackage->getName() == name && modulePackage->getStream() == stream) {
@@ -587,6 +663,7 @@ void ModulePackageContainer::install(const ModulePackage * module, const std::st
 void ModulePackageContainer::uninstall(const std::string &name, const std::string &stream,
     const std::string &profile)
 {
+    pImpl->addVersion2Modules();
     for (const auto &iter : pImpl->modules) {
         auto modulePackage = iter.second.get();
         if (modulePackage->getName() == name && modulePackage->getStream() == stream) {
@@ -617,44 +694,53 @@ ModulePackageContainer::Impl::moduleSolve(const std::vector<ModulePackage *> & m
     for (const auto &module : modules) {
         std::ostringstream ss;
         auto name = module->getName();
-        ss << "module(" << name << ":" << module->getStream() << ":" << module->getVersion() << ")";
+        ss << "module(" << name << ":" << module->getStream() << ")";
         Selector selector(moduleSack);
         bool optional = persistor->getState(name) == ModuleState::DEFAULT;
         selector.set(HY_PKG_PROVIDES, HY_EQ, ss.str().c_str());
         goal.install(&selector, optional);
         goalWeak.install(&selector, true);
     }
-    auto ret = goal.run(DNF_IGNORE_WEAK);
+    auto ret = goal.run(static_cast<DnfGoalActions>(DNF_IGNORE_WEAK | DNF_FORCE_BEST));
     if (debugSolver) {
         goal.writeDebugdata("debugdata/modules");
     }
     std::vector<std::vector<std::string>> problems;
     auto problemType = ModulePackageContainer::ModuleErrorType::NO_ERROR;
     if (ret) {
+        // Goal run ignor problem in defaults
         problems = goal.describeAllProblemRules(false);
-        ret = goal.run(DNF_NONE);
+        ret = goal.run(DNF_FORCE_BEST);
         if (ret) {
-            // Conflicting modules has to be removed otherwice it could result than one of them will
-            // be active
-            auto conflictingPkgs = goal.listConflictPkgs(DNF_PACKAGE_STATE_AVAILABLE);
-            dnf_sack_add_excludes(moduleSack, conflictingPkgs.get());
-            ret = goalWeak.run(DNF_NONE);
+            // Goal run ignor problem in defaults and in latest
+            ret = goal.run(DNF_NONE);
             if (ret) {
-                auto logger(Log::getLogger());
-                logger->critical("Modularity filtering totally broken\n");
-                problemType = ModulePackageContainer::ModuleErrorType::CANNOT_RESOLVE_MODULES;
-                activatedModules.reset();
-            } else {
-                problemType = ModulePackageContainer::ModuleErrorType::ERROR;
-                activatedModules.reset(new PackageSet(std::move(goalWeak.listInstalls())));
+                // Conflicting modules has to be removed otherwice it could result than one of them will
+                // be active
+                auto conflictingPkgs = goal.listConflictPkgs(DNF_PACKAGE_STATE_AVAILABLE);
+                dnf_sack_add_excludes(moduleSack, conflictingPkgs.get());
+                ret = goalWeak.run(DNF_NONE);
+                if (ret) {
+                    auto logger(Log::getLogger());
+                    logger->critical("Modularity filtering totally broken\n");
+                    problemType = ModulePackageContainer::ModuleErrorType::CANNOT_RESOLVE_MODULES;
+                    activatedModules.reset();
+                } else {
+                    problemType = ModulePackageContainer::ModuleErrorType::ERROR;
+                    Query query(moduleSack, Query::ExcludeFlags::IGNORE_EXCLUDES);
+                    goal2name_query(goalWeak, query);
+                    activatedModules.reset(new PackageSet(*query.runSet()));
+                }
+                return make_pair(problems, problemType);
             }
+            problemType = ModulePackageContainer::ModuleErrorType::ERROR_IN_LATEST;
         } else {
             problemType = ModulePackageContainer::ModuleErrorType::ERROR_IN_DEFAULTS;
-            activatedModules.reset(new PackageSet(std::move(goal.listInstalls())));
         }
-    } else {
-        activatedModules.reset(new PackageSet(std::move(goal.listInstalls())));
     }
+    Query query(moduleSack, Query::ExcludeFlags::IGNORE_EXCLUDES);
+    goal2name_query(goal, query);
+    activatedModules.reset(new PackageSet(*query.runSet()));
     return make_pair(problems, problemType);
 }
 
@@ -668,6 +754,7 @@ ModulePackageContainer::query(Nsvcap& moduleNevra)
 std::vector<ModulePackage *>
 ModulePackageContainer::query(std::string subject)
 {
+    pImpl->addVersion2Modules();
     // Alternatively a search using module provides could be performed
     std::vector<ModulePackage *> result;
     Query query(pImpl->moduleSack, Query::ExcludeFlags::IGNORE_EXCLUDES);
@@ -688,19 +775,26 @@ std::vector<ModulePackage *>
 ModulePackageContainer::query(std::string name, std::string stream, std::string version,
     std::string context, std::string arch)
 {
+    pImpl->addVersion2Modules();
     // Alternatively a search using module provides could be performed
     std::vector<ModulePackage *> result;
     Query query(pImpl->moduleSack, Query::ExcludeFlags::IGNORE_EXCLUDES);
     // platform modules are installed and not in modules std::Map.
     query.available();
-    std::ostringstream ss;
-    ss << stringFormater(name) << ":" << stringFormater(stream);
-    ss << ":" << stringFormater(version) << ":";
-    ss << stringFormater(context);
-    query.addFilter(HY_PKG_NAME, HY_GLOB, ss.str().c_str());
+    if (!name.empty() || !stream.empty()) {
+        std::ostringstream ss;
+        ss << stringFormater(name) << ":" << stringFormater(stream);
+        query.addFilter(HY_PKG_DESCRIPTION, HY_GLOB, ss.str().c_str());
+    }
+    if (!context.empty()) {
+        query.addFilter(HY_PKG_SUMMARY, HY_GLOB, context.c_str());
+    }
     if (!arch.empty()) {
         query.addFilter(HY_PKG_ARCH, HY_GLOB, arch.c_str());
     }
+    if (!version.empty()) {
+        query.addFilter(HY_PKG_VERSION, HY_GLOB, version.c_str());
+    }
     auto pset = query.runSet();
     Id moduleId = -1;
     while ((moduleId = pset->next(moduleId)) != -1) {
@@ -723,11 +817,11 @@ void ModulePackageContainer::enableDependencyTree(std::vector<ModulePackage *> &
         Query query(pImpl->moduleSack);
         query.addFilter(HY_PKG, HY_EQ, pImpl->activatedModules.get());
         auto pkg = dnf_package_new(pImpl->moduleSack, modulePackage->getId());
-        auto requires = dnf_package_get_requires(pkg);
-        query.addFilter(HY_PKG_PROVIDES, requires);
+        auto dep_requires = dnf_package_get_requires(pkg);
+        query.addFilter(HY_PKG_PROVIDES, dep_requires);
         auto set = query.runSet();
         toEnable += *set;
-        delete requires;
+        delete dep_requires;
         g_object_unref(pkg);
         enable(modulePackage);
         enabled.set(modulePackage->getId());
@@ -742,11 +836,11 @@ void ModulePackageContainer::enableDependencyTree(std::vector<ModulePackage *> &
             query.addFilter(HY_PKG, HY_EQ, pImpl->activatedModules.get());
             query.addFilter(HY_PKG, HY_NEQ, &enabled);
             auto pkg = dnf_package_new(pImpl->moduleSack, moduleId);
-            auto requires = dnf_package_get_requires(pkg);
-            query.addFilter(HY_PKG_PROVIDES, requires);
+            auto dep_requires = dnf_package_get_requires(pkg);
+            query.addFilter(HY_PKG_PROVIDES, dep_requires);
             auto set = query.runSet();
             toEnable += *set;
-            delete requires;
+            delete dep_requires;
             g_object_unref(pkg);
         }
         toEnable -= enabled;
@@ -763,8 +857,27 @@ ModulePackageContainer::getModuleState(const std::string& name)
     }
 }
 
+ModulePackage * ModulePackageContainer::getLatestModule(std::vector<ModulePackage *> modulePackages, bool activeOnly)
+{
+    ModulePackage * latest = nullptr;
+    for (ModulePackage * module: modulePackages) {
+        if (!activeOnly || isModuleActive(module->getId())) {
+            if (!latest) {
+                latest = module;
+            } else {
+                if (module->getVersionNum() > latest->getVersionNum()) {
+                    latest = module;
+                }
+            }
+        }
+    }
+
+    return latest;
+}
+
 std::set<std::string> ModulePackageContainer::getInstalledPkgNames()
 {
+    pImpl->addVersion2Modules();
     auto moduleNames = pImpl->persistor->getAllModuleNames();
     std::set<std::string> pkgNames;
     for (auto & moduleName: moduleNames) {
@@ -780,28 +893,9 @@ std::set<std::string> ModulePackageContainer::getInstalledPkgNames()
         nameStream += ":";
         nameStream += stream;
         auto modules = query(nameStream);
-        const ModulePackage * latest = nullptr;
-        for (const ModulePackage * module: modules) {
-            if (isModuleActive(module->getId())) {
-                if (!latest) {
-                    latest = module;
-                } else {
-                    if (module->getVersion() > latest->getVersion()) {
-                        latest = module;
-                    }
-                }
-            }
-        }
+        const ModulePackage * latest = getLatestModule(modules, true);
         if (!latest) {
-            for (auto module: modules) {
-                if (!latest) {
-                    latest = module;
-                } else {
-                    if (module->getVersion() > latest->getVersion()) {
-                        latest = module;
-                    }
-                }
-            }
+            latest = getLatestModule(modules, false);
         }
         if (!latest) {
             continue;
@@ -927,25 +1021,11 @@ modulePackageLatestPerRepoSorter(DnfSack * sack, const ModulePackage * first, co
     return first->getVersionNum() > second->getVersionNum();
 }
 
-static bool
-modulePackageLatestSorter(DnfSack * sack, const ModulePackage * first, const ModulePackage * second)
-{
-    int cmp = g_strcmp0(first->getNameCStr(), second->getNameCStr());
-    if (cmp != 0)
-        return cmp < 0;
-    cmp = dnf_sack_evr_cmp(sack, first->getStreamCStr(), second->getStreamCStr());
-    if (cmp != 0)
-        return cmp < 0;
-    cmp = g_strcmp0(first->getArchCStr(), second->getArchCStr());
-    if (cmp != 0)
-        return cmp < 0;
-    return first->getVersionNum() > second->getVersionNum();
-}
-
 std::vector<std::vector<std::vector<ModulePackage *>>>
 ModulePackageContainer::getLatestModulesPerRepo(ModuleState moduleFilter,
     std::vector<ModulePackage *> modulePackages)
 {
+    pImpl->addVersion2Modules();
     if (modulePackages.empty()) {
         return {};
     }
@@ -977,13 +1057,15 @@ ModulePackageContainer::getLatestModulesPerRepo(ModuleState moduleFilter,
     if (modulePackages.empty()) {
         return {};
     }
-    auto & packageFirst = modulePackages[0];
+
     std::vector<std::vector<std::vector<ModulePackage *>>> output;
     auto sack = pImpl->moduleSack;
     std::sort(modulePackages.begin(), modulePackages.end(),
               [sack](const ModulePackage * first, const ModulePackage * second)
               {return modulePackageLatestPerRepoSorter(sack, first, second);});
     auto vectorSize = modulePackages.size();
+
+    auto & packageFirst = modulePackages[0];
     output.push_back(
         std::vector<std::vector<ModulePackage *>>{std::vector<ModulePackage *> {packageFirst}});
     int repoIndex = 0;
@@ -1025,10 +1107,40 @@ ModulePackageContainer::getLatestModulesPerRepo(ModuleState moduleFilter,
     return output;
 }
 
+std::vector<ModulePackage *>
+ModulePackageContainer::getLatestModules(const std::vector<ModulePackage *> modulePackages, bool activeOnly)
+{
+    // Because modular sovables uses as name combination of module $name:$stream:$context, we can use to get the lates
+    // Query
+    std::vector<ModulePackage *> latestModules;
+    Query query(pImpl->moduleSack, Query::ExcludeFlags::IGNORE_EXCLUDES);
+    if (activeOnly) {
+        // When no active module return
+        if (!pImpl->activatedModules) {
+            return latestModules;
+        }
+        query.addFilter(HY_PKG, HY_EQ, pImpl->activatedModules.get());
+    }
+
+    PackageSet inputModulePackages(pImpl->moduleSack);
+    for (auto modulePackage : modulePackages) {
+        inputModulePackages.set(modulePackage->getId());
+    }
+    query.addFilter(HY_PKG, HY_EQ, &inputModulePackages);
+    query.addFilter(HY_PKG_LATEST_PER_ARCH, HY_EQ, 1);
+    auto set = query.runSet();
+
+    Id moduleId = -1;
+    while ((moduleId = set->next(moduleId)) != -1) {
+        latestModules.push_back(pImpl->modules.at(moduleId).get());
+    }
+    return latestModules;
+}
 
 std::pair<std::vector<std::vector<std::string>>, ModulePackageContainer::ModuleErrorType>
 ModulePackageContainer::resolveActiveModulePackages(bool debugSolver)
 {
+    pImpl->addVersion2Modules();
     dnf_sack_reset_excludes(pImpl->moduleSack);
     std::vector<ModulePackage *> packages;
 
@@ -1079,6 +1191,7 @@ bool ModulePackageContainer::isModuleActive(const ModulePackage * modulePackage)
 
 std::vector<ModulePackage *> ModulePackageContainer::getModulePackages()
 {
+    pImpl->addVersion2Modules();
     std::vector<ModulePackage *> values;
     const auto & modules = pImpl->modules;
     std::transform(
@@ -1100,47 +1213,56 @@ void ModulePackageContainer::rollback()
 
 std::map<std::string, std::string> ModulePackageContainer::getEnabledStreams()
 {
+    pImpl->addVersion2Modules();
     return pImpl->persistor->getEnabledStreams();
 }
 
 std::vector<std::string> ModulePackageContainer::getDisabledModules()
 {
+    pImpl->addVersion2Modules();
     return pImpl->persistor->getDisabledModules();
 }
 
 std::map<std::string, std::string> ModulePackageContainer::getDisabledStreams()
 {
+    pImpl->addVersion2Modules();
     return pImpl->persistor->getDisabledStreams();
 }
 
 std::vector<std::string> ModulePackageContainer::getResetModules()
 {
+    pImpl->addVersion2Modules();
     return pImpl->persistor->getResetModules();
 }
 
 std::map<std::string, std::string> ModulePackageContainer::getResetStreams()
 {
+    pImpl->addVersion2Modules();
     return pImpl->persistor->getResetStreams();
 }
 
 std::map<std::string, std::pair<std::string, std::string>>
 ModulePackageContainer::getSwitchedStreams()
 {
+    pImpl->addVersion2Modules();
     return pImpl->persistor->getSwitchedStreams();
 }
 
 std::map<std::string, std::vector<std::string>> ModulePackageContainer::getInstalledProfiles()
 {
+    pImpl->addVersion2Modules();
     return pImpl->persistor->getInstalledProfiles();
 }
 
 std::vector<std::string> ModulePackageContainer::getInstalledProfiles(std::string moduleName)
 {
+    pImpl->addVersion2Modules();
     return pImpl->persistor->getProfiles(moduleName);
 }
 
 std::map<std::string, std::vector<std::string>> ModulePackageContainer::getRemovedProfiles()
 {
+    pImpl->addVersion2Modules();
     return pImpl->persistor->getRemovedProfiles();
 }
 const std::string &
@@ -1178,7 +1300,7 @@ ModulePackageContainer::Impl::ModulePersistor::changeStream(const std::string &n
     if (updatedValue == stream)
         return false;
     const auto &originValue = configs.at(name).first.getValue(name, "stream");
-    if (originValue != updatedValue) {
+    if (originValue != updatedValue && configs.at(name).second.streamChangesNum > 1) {
         throw EnableMultipleStreamsException(name);
     }
     getEntry(name).second.stream = stream;
@@ -1248,10 +1370,10 @@ static inline void
 parseConfig(ConfigParser &parser, const std::string &name, const char *path)
 {
     auto logger(Log::getLogger());
+    const auto fname = name + ".module";
+    g_autofree gchar * cfn = g_build_filename(path, fname.c_str(), NULL);
 
     try {
-        const auto fname = name + ".module";
-        g_autofree gchar * cfn = g_build_filename(path, fname.c_str(), NULL);
         parser.read(cfn);
 
         /* FIXME: init empty config or throw error? */
@@ -1271,10 +1393,15 @@ parseConfig(ConfigParser &parser, const std::string &name, const char *path)
             parser.setValue(name, "state", parser.getValue(name, "enabled"));
             parser.removeOption(name, "enabled");
         }
-    } catch (const ConfigParser::CantOpenFile &) {
+    } catch (const ConfigParser::FileDoesNotExist &) {
         /* No module config file present. Fill values in */
         initConfig(parser, name);
         return;
+    } catch (const ConfigParser::CantOpenFile &) {
+        /* File exists but is not readable. */
+        logger->warning(tfm::format("Cannot read \"%s\". Modular filtering may be affected.", cfn));
+        initConfig(parser, name);
+        return;
     }
 }
 
@@ -1298,6 +1425,7 @@ bool ModulePackageContainer::Impl::ModulePersistor::insert(
 
     newConfig.state = fromString(parser.getValue(moduleName, "state"));
     newConfig.stream = parser.getValue(moduleName, "stream");
+    newConfig.streamChangesNum = 0;
 
     return true;
 }
@@ -1544,8 +1672,9 @@ ModulePackageContainer::Impl::ModulePersistor::getRemovedProfiles()
 
 void ModulePackageContainer::loadFailSafeData()
 {
+    pImpl->addVersion2Modules();
     auto persistor = pImpl->persistor->configs;
-    
+
     std::map<std::string, std::pair<std::string, bool>> enabledStreams;
     for (auto & nameConfig: persistor) {
         if (nameConfig.second.second.state == ModuleState::ENABLED) {
@@ -1598,114 +1727,157 @@ void ModulePackageContainer::loadFailSafeData()
 
 std::vector<ModulePackage *> ModulePackageContainer::Impl::getLatestActiveEnabledModules()
 {
+    Query query(moduleSack, Query::ExcludeFlags::IGNORE_EXCLUDES);
+    query.addFilter(HY_PKG, HY_EQ, activatedModules.get());
+    query.addFilter(HY_PKG_REPONAME, HY_NEQ, HY_SYSTEM_REPO_NAME);
+    query.addFilter(HY_PKG_LATEST_PER_ARCH, HY_EQ, 1);
+    auto set = query.runSet();
+
     std::vector<ModulePackage *> activeModules;
     Id moduleId = -1;
-    while ((moduleId = activatedModules->next(moduleId)) != -1) {
+    while ((moduleId = set->next(moduleId)) != -1) {
         auto modulePackage = modules.at(moduleId).get();
         if (isEnabled(modulePackage->getName(), modulePackage->getStream())) {
             activeModules.push_back(modulePackage);
         }
     }
-    if (activeModules.empty()) {
-        return {};
+    return activeModules;
+}
+
+void ModulePackageContainer::Impl::addVersion2Modules()
+{
+    if (modulesV2.empty()) {
+        return;
     }
-    auto sack = moduleSack;
-    std::sort(activeModules.begin(), activeModules.end(),
-              [sack](const ModulePackage * first, const ModulePackage * second)
-              {return modulePackageLatestSorter(sack, first, second);});
-
-    auto packageFirst = activeModules[0];
-    std::vector<ModulePackage *> latest;
-    auto vectorSize = activeModules.size();
-    latest.push_back(packageFirst);
-    auto name = packageFirst->getNameCStr();
-    auto stream = packageFirst->getStreamCStr();
-    auto arch = packageFirst->getArchCStr();
-
-    for (unsigned int index = 1; index < vectorSize; ++index) {
-        auto & package = activeModules[index];
-        if (g_strcmp0(package->getNameCStr(), name) != 0 ||
-            g_strcmp0(package->getStreamCStr(), stream) != 0 ||
-            g_strcmp0(package->getArchCStr(), arch) != 0) {
-            name = package->getNameCStr();
-            stream = package->getStreamCStr();
-            arch = package->getArchCStr();
-            latest.push_back(package);
+    std::map<std::string, std::map<std::string, std::vector<ModulePackage *>>> v3_context_map;
+    for (auto const & module_pair : modules) {
+        auto * module = module_pair.second.get();
+        auto dep_requires = module->getRequires(true);
+        auto concentratedRequires = concentrateVectorString(dep_requires);
+        v3_context_map[module->getNameStream()][concentratedRequires].push_back(module);
+    }
+    libdnf::LibsolvRepo * repo;
+    ModulemdModuleStream * mdStream;
+    std::string repoID;
+    g_autofree gchar * path = g_build_filename(installRoot.c_str(), "/etc/dnf/modules.d", NULL);
+    for (auto & module_tuple : modulesV2) {
+        std::tie(repo, mdStream, repoID) = module_tuple;
+        auto nameStream = ModulePackage::getNameStream(mdStream);
+        auto dep_requires = ModulePackage::getRequires(mdStream, true);
+        auto concentratedRequires = concentrateVectorString(dep_requires);
+        auto streamIterator = v3_context_map.find(nameStream);
+        if (streamIterator != v3_context_map.end()) {
+            auto contextIterator = streamIterator->second.find(concentratedRequires);
+            if (contextIterator != streamIterator->second.end()) {
+                auto v3_context = contextIterator->second[0]->getContext();
+                std::unique_ptr<ModulePackage> modulePackage(new ModulePackage(moduleSack, repo, mdStream, repoID, v3_context));
+                persistor->insert(modulePackage->getName(), path);
+                modules.insert(std::make_pair(modulePackage->getId(), std::move(modulePackage)));
+                g_object_unref(mdStream);
+                continue;
+            }
         }
+        if (concentratedRequires.empty()) {
+            concentratedRequires.append("NoRequires");
+        }
+        std::unique_ptr<ModulePackage> modulePackage(new ModulePackage(moduleSack, repo, mdStream, repoID, concentratedRequires));
+        persistor->insert(modulePackage->getName(), path);
+        modules.insert(std::make_pair(modulePackage->getId(), std::move(modulePackage)));
+        g_object_unref(mdStream);
     }
-    return latest;
+    modulesV2.clear();
 }
 
 void ModulePackageContainer::updateFailSafeData()
 {
-    auto activatedModules = pImpl->activatedModules.get();
-    if (!activatedModules) {
-        return;
-    }
-    std::vector<ModulePackage *> latest = pImpl->getLatestActiveEnabledModules();
-
     auto fileNames = getYamlFilenames(pImpl->persistDir.c_str());
-    Map map;
-    map_init(&map, fileNames.size());
-    auto begin = fileNames.begin();
-    auto end = fileNames.end();
-    if (g_mkdir_with_parents(pImpl->persistDir.c_str(), 0755) == -1) {
-        const char * errTxt = strerror(errno);
-        auto logger(Log::getLogger());
-        logger->debug(tfm::format(
-            _("Unable to create directory \"%s\" for modular Fail Safe data: %s"),
-            pImpl->persistDir.c_str(), errTxt));
-    }
 
-    // Update FailSafe data
-    for (auto modulePackage: latest) {
-        std::ostringstream ss;
-        ss << modulePackage->getNameStream();
-        ss << ":" << modulePackage->getArch() << ".yaml";
-        auto fileName = ss.str();
-        auto low = std::lower_bound(begin, end, fileName);
-        if (low != end && fileName == *low) {
-            MAPSET(&map, low - begin);
-        }
-        if (modulePackage->getRepoID() == LIBDNF_MODULE_FAIL_SAFE_REPO_NAME) {
-            continue;
-        }
-        g_autofree gchar * filePath = g_build_filename(pImpl->persistDir.c_str(), fileName.c_str(),
-                                                       NULL);
-        if (!updateFile(filePath, modulePackage->getYaml().c_str())) {
+    if (pImpl->activatedModules) {
+        std::vector<ModulePackage *> latest = pImpl->getLatestActiveEnabledModules();
+
+        if (g_mkdir_with_parents(pImpl->persistDir.c_str(), 0755) == -1) {
+            const char * errTxt = strerror(errno);
             auto logger(Log::getLogger());
             logger->debug(tfm::format(
-                _("Unable to save a modular Fail Safe data to '%s'"), filePath));
+                _("Unable to create directory \"%s\" for modular Fail Safe data: %s"),
+                                      pImpl->persistDir.c_str(), errTxt));
+        }
+
+        // Update FailSafe data
+        for (auto modulePackage: latest) {
+            std::ostringstream ss;
+            ss << modulePackage->getNameStream();
+            ss << ":" << modulePackage->getArch() << ".yaml";
+            auto fileName = ss.str();
+            if (modulePackage->getRepoID() == LIBDNF_MODULE_FAIL_SAFE_REPO_NAME) {
+                continue;
+            }
+            g_autofree gchar * filePath = g_build_filename(pImpl->persistDir.c_str(), fileName.c_str(), NULL);
+            if (!updateFile(filePath, modulePackage->getYaml().c_str())) {
+                auto logger(Log::getLogger());
+                logger->debug(tfm::format(_("Unable to save a modular Fail Safe data to '%s'"), filePath));
+            }
         }
     }
 
     // Remove files from not enabled modules
     for (unsigned int index = 0; index < fileNames.size(); ++index) {
-        if (!MAPTST(&map, index)) {
-            auto fileName = fileNames[index];
-            auto first = fileName.find(":");
-            if (first == std::string::npos || first == 0) {
-                continue;
-            }
-            std::string moduleName = fileName.substr(0, first);
-            auto second = fileName.find(":", ++first);
-            if (second == std::string::npos || first == second) {
-                continue;
-            }
-            std::string moduleStream = fileName.substr(first, second - first);
+        auto fileName = fileNames[index];
+        auto first = fileName.find(":");
+        if (first == std::string::npos || first == 0) {
+            continue;
+        }
+        std::string moduleName = fileName.substr(0, first);
+        auto second = fileName.find(":", ++first);
+        if (second == std::string::npos || first == second) {
+            continue;
+        }
+        std::string moduleStream = fileName.substr(first, second - first);
 
-            if (!isEnabled(moduleName, moduleStream)) {
-                g_autofree gchar * file = g_build_filename(
-                    pImpl->persistDir.c_str(), fileNames[index].c_str(), NULL);
-                if (remove(file)) {
+        if (!isEnabled(moduleName, moduleStream)) {
+            g_autofree gchar * file = g_build_filename(pImpl->persistDir.c_str(), fileNames[index].c_str(), NULL);
+            if (remove(file)) {
+                auto logger(Log::getLogger());
+                logger->debug(tfm::format(_("Unable to remove a modular Fail Safe data in '%s'"), file));
+            }
+        }
+    }
+}
+
+void ModulePackageContainer::applyObsoletes(){
+    for (const auto &iter : pImpl->modules) {
+        auto modulePkg = iter.second.get();
+        if (!isEnabled(modulePkg)) {
+            continue;
+        }
+        /* We cannot access the eol through the ModulemdModuleStream which is
+         * wrapped in modulePkg because it was created with metedata from only
+         * one repository but other repositories can have new eols which affect
+         * this module stream. We have to use merged modular metadata from all
+         * available repositories.
+         */
+        ModulemdObsoletes *modulePkgObsoletes = pImpl->moduleMetadata.getNewestActiveObsolete(modulePkg);
+        if (modulePkgObsoletes) {
+            auto moduleName = modulemd_obsoletes_get_obsoleted_by_module_name(modulePkgObsoletes);
+            auto moduleStream = modulemd_obsoletes_get_obsoleted_by_module_stream(modulePkgObsoletes);
+
+            if (moduleName && moduleStream) {
+                if (!isDisabled(moduleName)) {
+                    enable(moduleName, moduleStream, false);
+                    if (std::string(moduleName) != modulePkg->getName()) {
+                        reset(modulePkg, false);
+                    }
+                } else {
                     auto logger(Log::getLogger());
                     logger->debug(tfm::format(
-                        _("Unable to remove a modular Fail Safe data in '%s'"), file));
+                        _("Unable to apply modular obsoletes to '%s:%s' because target module '%s' is disabled"),
+                        modulePkg->getName(), modulePkg->getStream(), moduleName));
                 }
+            } else {
+                reset(modulePkg, false);
             }
         }
     }
-    map_free(&map);
 }
 
 }
