@@ -40,7 +40,7 @@
 
 #include "bgettext/bgettext-lib.h"
 #include "tinyformat/tinyformat.hpp"
-#include <utils.hpp>
+#include "utils.hpp"
 
 #include <librepo/librepo.h>
 #include <fcntl.h>
@@ -49,8 +49,6 @@
 #include <unistd.h>
 #include <utime.h>
 
-#include <gpgme.h>
-
 #include <solv/chksum.h>
 #include <solv/repo.h>
 #include <solv/util.h>
@@ -112,11 +110,6 @@ struct default_delete<LrPackageTarget> {
     void operator()(LrPackageTarget * ptr) noexcept { lr_packagetarget_free(ptr); }
 };
 
-template<>
-struct default_delete<std::remove_pointer<gpgme_ctx_t>::type> {
-    void operator()(gpgme_ctx_t ptr) noexcept { gpgme_release(ptr); }
-};
-
 } // namespace std
 
 namespace libdnf {
@@ -200,7 +193,7 @@ const std::string & Repo::Impl::getMetadataPath(const std::string &metadataType)
 //    auto logger(Log::getLogger());
     static const std::string empty;
     std::string lookupMetadataType = metadataType;
-    if (conf->getMasterConfig().zchunk().getValue()) {
+    if (conf->getMainConfig().zchunk().getValue()) {
         if(!endsWith(metadataType, "_zck"))
             lookupMetadataType = metadataType + "_zck";
     }
@@ -259,20 +252,16 @@ int Repo::Impl::mirrorFailureCB(void * data, const char * msg, const char * url,
 /**
 * @brief Format user password string
 *
-* Returns user and password in user:password form. If quote is True,
-* special characters in user and password are URL encoded.
+* Returns user and password in user:password form.
+* Special characters in user and password are URL encoded.
 *
 * @param user Username
 * @param passwd Password
-* @param encode If quote is True, special characters in user and password are URL encoded.
 * @return User and password in user:password form
 */
-static std::string formatUserPassString(const std::string & user, const std::string & passwd, bool encode)
+static std::string formatUserPassString(const std::string & user, const std::string & passwd)
 {
-    if (encode)
-        return urlEncode(user) + ":" + urlEncode(passwd);
-    else
-        return user + ":" + passwd;
+    return urlEncode(user) + ":" + urlEncode(passwd);
 }
 
 Repo::Impl::Impl(Repo & owner, const std::string & id, Type type, std::unique_ptr<ConfigRepo> && conf)
@@ -379,7 +368,7 @@ std::string Repo::getLocalBaseurl() const
 }
 
 bool Repo::load() { return pImpl->load(); }
-bool Repo::loadCache(bool throwExcept) { return pImpl->loadCache(throwExcept); }
+bool Repo::loadCache(bool throwExcept, bool ignoreMissing) { return pImpl->loadCache(throwExcept, ignoreMissing); }
 void Repo::downloadMetadata(const std::string & destdir) { pImpl->downloadMetadata(destdir); }
 bool Repo::getUseIncludes() const { return pImpl->useIncludes; }
 void Repo::setUseIncludes(bool enabled) { pImpl->useIncludes = enabled; }
@@ -439,12 +428,18 @@ std::string Repo::getMetadataContent(const std::string &metadataType)
 std::unique_ptr<LrHandle> Repo::Impl::lrHandleInitBase()
 {
     std::unique_ptr<LrHandle> h(lr_handle_init());
-    std::vector<const char *> dlist = {MD_TYPE_PRIMARY, MD_TYPE_FILELISTS, MD_TYPE_PRESTODELTA,
-        MD_TYPE_GROUP_GZ, MD_TYPE_UPDATEINFO};
+    std::vector<const char *> dlist = {MD_TYPE_PRIMARY, MD_TYPE_PRESTODELTA, MD_TYPE_GROUP_GZ, MD_TYPE_UPDATEINFO};
+
+    auto & optionalMetadataTypes = conf->getMainConfig().optional_metadata_types().getValue();
+    auto loadFilelists = std::find(optionalMetadataTypes.begin(), optionalMetadataTypes.end(), "filelists") !=
+                         optionalMetadataTypes.end();
 
 #ifdef MODULEMD
     dlist.push_back(MD_TYPE_MODULES);
 #endif
+    if (loadFilelists) {
+        dlist.push_back(MD_TYPE_FILELISTS);
+    }
     if (loadMetadataOther) {
         dlist.push_back(MD_TYPE_OTHER);
     }
@@ -461,6 +456,8 @@ std::unique_ptr<LrHandle> Repo::Impl::lrHandleInitBase()
     handleSetOpt(h.get(), LRO_MAXMIRRORTRIES, static_cast<long>(maxMirrorTries));
     handleSetOpt(h.get(), LRO_MAXPARALLELDOWNLOADS,
                      conf->max_parallel_downloads().getValue());
+    handleSetOpt(h.get(), LRO_MAXDOWNLOADSPERMIRROR,
+                     conf->max_downloads_per_mirror().getValue());
 
     LrUrlVars * vars = NULL;
     vars = lr_urlvars_set(vars, MD_TYPE_GROUP_GZ, MD_TYPE_GROUP);
@@ -484,7 +481,7 @@ std::unique_ptr<LrHandle> Repo::Impl::lrHandleInitLocal()
     handleSetOpt(h.get(), LRO_LOCAL, 1L);
 #ifdef LRO_SUPPORTS_CACHEDIR
     /* If zchunk is enabled, set librepo cache dir */
-    if (conf->getMasterConfig().zchunk().getValue()) {
+    if (conf->getMainConfig().zchunk().getValue()) {
         if (conf->basecachedir().empty()) {
             throw Exception(tfm::format(_("repo '%s': 'basecachedir' is not set"), id));
         }
@@ -494,6 +491,84 @@ std::unique_ptr<LrHandle> Repo::Impl::lrHandleInitLocal()
     return h;
 }
 
+template<typename ConfigT>
+static void setHandle(LrHandle * h, ConfigT & config, const char * repoId = nullptr) {
+    auto & ipResolve = config.ip_resolve().getValue();
+    if (ipResolve == "ipv4")
+        handleSetOpt(h, LRO_IPRESOLVE, LR_IPRESOLVE_V4);
+    else if (ipResolve == "ipv6")
+        handleSetOpt(h, LRO_IPRESOLVE, LR_IPRESOLVE_V6);
+
+    auto minrate = config.minrate().getValue();
+    handleSetOpt(h, LRO_LOWSPEEDLIMIT, static_cast<long>(minrate));
+
+    auto maxspeed = config.throttle().getValue();
+    if (maxspeed > 0 && maxspeed <= 1)
+        maxspeed *= config.bandwidth().getValue();
+    if (maxspeed != 0 && maxspeed < minrate)
+        throw RepoError(_("Maximum download speed is lower than minimum. "
+                          "Please change configuration of minrate or throttle"));
+    handleSetOpt(h, LRO_MAXSPEED, static_cast<int64_t>(maxspeed));
+
+    long timeout = config.timeout().getValue();
+    if (timeout > 0) {
+        handleSetOpt(h, LRO_CONNECTTIMEOUT, timeout);
+        handleSetOpt(h, LRO_LOWSPEEDTIME, timeout);
+    } else {
+        handleSetOpt(h, LRO_CONNECTTIMEOUT, LRO_CONNECTTIMEOUT_DEFAULT);
+        handleSetOpt(h, LRO_LOWSPEEDTIME, LRO_LOWSPEEDTIME_DEFAULT);
+    }
+
+    // setup username/password if needed
+    handleSetOpt(h, LRO_USERNAME, config.username().getValue().empty() ? NULL : config.username().getValue().c_str());
+    handleSetOpt(h, LRO_PASSWORD, config.password().getValue().empty() ? NULL : config.password().getValue().c_str());
+
+    if (!config.proxy().empty() && !config.proxy().getValue().empty())
+        handleSetOpt(h, LRO_PROXY, config.proxy().getValue().c_str());
+
+    //set proxy authorization method
+    auto proxyAuthMethods = Repo::Impl::stringToProxyAuthMethods(config.proxy_auth_method().getValue());
+    handleSetOpt(h, LRO_PROXYAUTHMETHODS, static_cast<long>(proxyAuthMethods));
+
+    // setup proxy username/password if needed
+    if (!config.proxy_username().empty()) {
+        auto userpwd = config.proxy_username().getValue();
+        if (!userpwd.empty()) {
+            if (config.proxy_password().empty()) {
+                if (repoId)
+                    throw RepoError(tfm::format(_("repo '%s': 'proxy_username' is set but not 'proxy_password'"), repoId));
+                else
+                    throw RepoError(_("'proxy_username' is set but not 'proxy_password'"));
+            }
+            userpwd = formatUserPassString(userpwd, config.proxy_password().getValue());
+            handleSetOpt(h, LRO_PROXYUSERPWD, userpwd.c_str());
+        }
+    }
+
+    // setup ssl stuff
+    if (!config.sslcacert().getValue().empty())
+        handleSetOpt(h, LRO_SSLCACERT, config.sslcacert().getValue().c_str());
+    if (!config.sslclientcert().getValue().empty())
+        handleSetOpt(h, LRO_SSLCLIENTCERT, config.sslclientcert().getValue().c_str());
+    if (!config.sslclientkey().getValue().empty())
+        handleSetOpt(h, LRO_SSLCLIENTKEY, config.sslclientkey().getValue().c_str());
+    long sslverify = config.sslverify().getValue() ? 1L : 0L;
+    handleSetOpt(h, LRO_SSLVERIFYHOST, sslverify);
+    handleSetOpt(h, LRO_SSLVERIFYPEER, sslverify);
+    handleSetOpt(h, LRO_SSLVERIFYSTATUS, config.sslverifystatus().getValue() ? 1L : 0L);
+
+    // setup proxy ssl stuff
+    if (!config.proxy_sslcacert().getValue().empty())
+        handleSetOpt(h, LRO_PROXY_SSLCACERT, config.proxy_sslcacert().getValue().c_str());
+    if (!config.proxy_sslclientcert().getValue().empty())
+        handleSetOpt(h, LRO_PROXY_SSLCLIENTCERT, config.proxy_sslclientcert().getValue().c_str());
+    if (!config.proxy_sslclientkey().getValue().empty())
+        handleSetOpt(h, LRO_PROXY_SSLCLIENTKEY, config.proxy_sslclientkey().getValue().c_str());
+    long proxy_sslverify = config.proxy_sslverify().getValue() ? 1L : 0L;
+    handleSetOpt(h, LRO_PROXY_SSLVERIFYHOST, proxy_sslverify);
+    handleSetOpt(h, LRO_PROXY_SSLVERIFYPEER, proxy_sslverify);
+}
+
 std::unique_ptr<LrHandle> Repo::Impl::lrHandleInitRemote(const char *destdir)
 {
     std::unique_ptr<LrHandle> h(lrHandleInitBase());
@@ -506,12 +581,6 @@ std::unique_ptr<LrHandle> Repo::Impl::lrHandleInitRemote(const char *destdir)
 
     handleSetOpt(h.get(), LRO_DESTDIR, destdir);
 
-    auto & ipResolve = conf->ip_resolve().getValue();
-    if (ipResolve == "ipv4")
-        handleSetOpt(h.get(), LRO_IPRESOLVE, LR_IPRESOLVE_V4);
-    else if (ipResolve == "ipv6")
-        handleSetOpt(h.get(), LRO_IPRESOLVE, LR_IPRESOLVE_V6);
-
     enum class Source {NONE, METALINK, MIRRORLIST} source{Source::NONE};
     std::string tmp;
     if (!conf->metalink().empty() && !(tmp=conf->metalink().getValue()).empty())
@@ -552,22 +621,6 @@ std::unique_ptr<LrHandle> Repo::Impl::lrHandleInitRemote(const char *destdir)
     if (source == Source::NONE && conf->baseurl().getValue().empty())
         throw RepoError(tfm::format(_("Cannot find a valid baseurl for repo: %s"), id));
 
-    // setup username/password if needed
-    auto userpwd = conf->username().getValue();
-    if (!userpwd.empty()) {
-        // TODO Use URL encoded form, needs support in librepo
-        userpwd = formatUserPassString(userpwd, conf->password().getValue(), false);
-        handleSetOpt(h.get(), LRO_USERPWD, userpwd.c_str());
-    }
-
-    // setup ssl stuff
-    if (!conf->sslcacert().getValue().empty())
-        handleSetOpt(h.get(), LRO_SSLCACERT, conf->sslcacert().getValue().c_str());
-    if (!conf->sslclientcert().getValue().empty())
-        handleSetOpt(h.get(), LRO_SSLCLIENTCERT, conf->sslclientcert().getValue().c_str());
-    if (!conf->sslclientkey().getValue().empty())
-        handleSetOpt(h.get(), LRO_SSLCLIENTKEY, conf->sslclientkey().getValue().c_str());
-
     handleSetOpt(h.get(), LRO_HMFCB, static_cast<LrHandleMirrorFailureCb>(mirrorFailureCB));
     handleSetOpt(h.get(), LRO_PROGRESSCB, static_cast<LrProgressCb>(progressCB));
     handleSetOpt(h.get(), LRO_PROGRESSDATA, callbacks.get());
@@ -576,7 +629,7 @@ std::unique_ptr<LrHandle> Repo::Impl::lrHandleInitRemote(const char *destdir)
 
 #ifdef LRO_SUPPORTS_CACHEDIR
     /* If zchunk is enabled, set librepo cache dir */
-    if (conf->getMasterConfig().zchunk().getValue()) {
+    if (conf->getMainConfig().zchunk().getValue()) {
         if (conf->basecachedir().empty()) {
             throw Exception(tfm::format(_("repo '%s': 'basecachedir' is not set"), id));
         }
@@ -584,217 +637,11 @@ std::unique_ptr<LrHandle> Repo::Impl::lrHandleInitRemote(const char *destdir)
     }
 #endif
 
-    auto minrate = conf->minrate().getValue();
-    handleSetOpt(h.get(), LRO_LOWSPEEDLIMIT, static_cast<long>(minrate));
-
-    auto maxspeed = conf->throttle().getValue();
-    if (maxspeed > 0 && maxspeed <= 1)
-        maxspeed *= conf->bandwidth().getValue();
-    if (maxspeed != 0 && maxspeed < minrate)
-        throw RepoError(_("Maximum download speed is lower than minimum. "
-                          "Please change configuration of minrate or throttle"));
-    handleSetOpt(h.get(), LRO_MAXSPEED, static_cast<int64_t>(maxspeed));
-
-    long timeout = conf->timeout().getValue();
-    if (timeout > 0) {
-        handleSetOpt(h.get(), LRO_CONNECTTIMEOUT, timeout);
-        handleSetOpt(h.get(), LRO_LOWSPEEDTIME, timeout);
-    } else {
-        handleSetOpt(h.get(), LRO_CONNECTTIMEOUT, LRO_CONNECTTIMEOUT_DEFAULT);
-        handleSetOpt(h.get(), LRO_LOWSPEEDTIME, LRO_LOWSPEEDTIME_DEFAULT);
-    }
-
-    if (!conf->proxy().empty() && !conf->proxy().getValue().empty())
-        handleSetOpt(h.get(), LRO_PROXY, conf->proxy().getValue().c_str());
-
-    //set proxy authorization method
-    auto proxyAuthMethodStr = conf->proxy_auth_method().getValue();
-    auto proxyAuthMethod = LR_AUTH_ANY;
-    for (auto & auth : PROXYAUTHMETHODS) {
-        if (proxyAuthMethodStr == auth.name) {
-            proxyAuthMethod = auth.code;
-            break;
-        }
-    }
-    handleSetOpt(h.get(), LRO_PROXYAUTHMETHODS, static_cast<long>(proxyAuthMethod));
-
-    if (!conf->proxy_username().empty()) {
-        userpwd = conf->proxy_username().getValue();
-        if (!userpwd.empty()) {
-            if (conf->proxy_password().empty()) {
-                throw RepoError(tfm::format(_("repo '%s': 'proxy_username' is set but not 'proxy_password'"), id));
-            }
-            userpwd = formatUserPassString(userpwd, conf->proxy_password().getValue(), true);
-            handleSetOpt(h.get(), LRO_PROXYUSERPWD, userpwd.c_str());
-        }
-    }
-
-    auto sslverify = conf->sslverify().getValue() ? 1L : 0L;
-    handleSetOpt(h.get(), LRO_SSLVERIFYHOST, sslverify);
-    handleSetOpt(h.get(), LRO_SSLVERIFYPEER, sslverify);
+    setHandle(h.get(), *conf, id.c_str());
 
     return h;
 }
 
-static void gpgImportKey(gpgme_ctx_t context, int keyFd)
-{
-    auto logger(Log::getLogger());
-    gpg_error_t gpgErr;
-    gpgme_data_t keyData;
-
-    gpgErr = gpgme_data_new_from_fd(&keyData, keyFd);
-    if (gpgErr != GPG_ERR_NO_ERROR) {
-        auto msg = tfm::format(_("%s: gpgme_data_new_from_fd(): %s"), __func__, gpgme_strerror(gpgErr));
-        logger->debug(msg);
-        throw LrException(LRE_GPGERROR, msg);
-    }
-
-    gpgErr = gpgme_op_import(context, keyData);
-    gpgme_data_release(keyData);
-    if (gpgErr != GPG_ERR_NO_ERROR) {
-        auto msg = tfm::format(_("%s: gpgme_op_import(): %s"), __func__, gpgme_strerror(gpgErr));
-        logger->debug(msg);
-        throw LrException(LRE_GPGERROR, msg);
-    }
-}
-
-static void gpgImportKey(gpgme_ctx_t context, std::vector<char> key)
-{
-    auto logger(Log::getLogger());
-    gpg_error_t gpgErr;
-    gpgme_data_t keyData;
-
-    gpgErr = gpgme_data_new_from_mem(&keyData, key.data(), key.size(), 0);
-    if (gpgErr != GPG_ERR_NO_ERROR) {
-        auto msg = tfm::format(_("%s: gpgme_data_new_from_fd(): %s"), __func__, gpgme_strerror(gpgErr));
-        logger->debug(msg);
-        throw LrException(LRE_GPGERROR, msg);
-    }
-
-    gpgErr = gpgme_op_import(context, keyData);
-    gpgme_data_release(keyData);
-    if (gpgErr != GPG_ERR_NO_ERROR) {
-        auto msg = tfm::format(_("%s: gpgme_op_import(): %s"), __func__, gpgme_strerror(gpgErr));
-        logger->debug(msg);
-        throw LrException(LRE_GPGERROR, msg);
-    }
-}
-
-static std::vector<Key> rawkey2infos(int fd) {
-    auto logger(Log::getLogger());
-    gpg_error_t gpgErr;
-
-    std::vector<Key> keyInfos;
-    gpgme_ctx_t ctx;
-    gpgme_new(&ctx);
-    std::unique_ptr<std::remove_pointer<gpgme_ctx_t>::type> context(ctx);
-
-    // set GPG home dir
-    char tmpdir[] = "/tmp/tmpdir.XXXXXX";
-    mkdtemp(tmpdir);
-    Finalizer tmpDirRemover([&tmpdir](){
-        dnf_remove_recursive(tmpdir, NULL);
-    });
-    gpgErr = gpgme_ctx_set_engine_info(ctx, GPGME_PROTOCOL_OpenPGP, NULL, tmpdir);
-    if (gpgErr != GPG_ERR_NO_ERROR) {
-        auto msg = tfm::format(_("%s: gpgme_ctx_set_engine_info(): %s"), __func__, gpgme_strerror(gpgErr));
-        logger->debug(msg);
-        throw LrException(LRE_GPGERROR, msg);
-    }
-
-    gpgImportKey(ctx, fd);
-
-    gpgme_key_t key;
-    gpgErr = gpgme_op_keylist_start(ctx, NULL, 0);
-    while (!gpgErr)
-    {
-        gpgErr = gpgme_op_keylist_next(ctx, &key);
-        if (gpgErr)
-            break;
-
-        // _extract_signing_subkey
-        auto subkey = key->subkeys;
-        while (subkey && !key->subkeys->can_sign) {
-            subkey = subkey->next;
-        }
-        if (subkey)
-            keyInfos.emplace_back(key, subkey);
-        gpgme_key_release(key);
-    }
-    if (gpg_err_code(gpgErr) != GPG_ERR_EOF)
-    {
-        gpgme_op_keylist_end(ctx);
-        auto msg = tfm::format(_("can not list keys: %s"), gpgme_strerror(gpgErr));
-        logger->debug(msg);
-        throw LrException(LRE_GPGERROR, msg);
-    }
-    gpgme_set_armor(ctx, 1);
-    for (auto & keyInfo : keyInfos) {
-        gpgme_data_t sink;
-        gpgme_data_new(&sink);
-        gpgme_op_export(ctx, keyInfo.getId().c_str(), 0, sink);
-        gpgme_data_rewind(sink);
-
-        char buf[4096];
-        ssize_t readed;
-        do {
-            readed = gpgme_data_read(sink, buf, sizeof(buf));
-            if (readed > 0)
-                keyInfo.rawKey.insert(keyInfo.rawKey.end(), buf, buf + sizeof(buf));
-        } while (readed == sizeof(buf));
-    }
-    return keyInfos;
-}
-
-static std::vector<std::string> keyidsFromPubring(const std::string & gpgDir)
-{
-    auto logger(Log::getLogger());
-    gpg_error_t gpgErr;
-
-    std::vector<std::string> keyids;
-
-    struct stat sb;
-    if (stat(gpgDir.c_str(), &sb) == 0 && S_ISDIR(sb.st_mode)) {
-
-        gpgme_ctx_t ctx;
-        gpgme_new(&ctx);
-        std::unique_ptr<std::remove_pointer<gpgme_ctx_t>::type> context(ctx);
-
-        // set GPG home dir
-        gpgErr = gpgme_ctx_set_engine_info(ctx, GPGME_PROTOCOL_OpenPGP, NULL, gpgDir.c_str());
-        if (gpgErr != GPG_ERR_NO_ERROR) {
-            auto msg = tfm::format(_("%s: gpgme_ctx_set_engine_info(): %s"), __func__, gpgme_strerror(gpgErr));
-            logger->debug(msg);
-            throw LrException(LRE_GPGERROR, msg);
-        }
-
-        gpgme_key_t key;
-        gpgErr = gpgme_op_keylist_start(ctx, NULL, 0);
-        while (!gpgErr)
-        {
-            gpgErr = gpgme_op_keylist_next(ctx, &key);
-            if (gpgErr)
-                break;
-
-            // _extract_signing_subkey
-            auto subkey = key->subkeys;
-            while (subkey && !key->subkeys->can_sign) {
-                subkey = subkey->next;
-            }
-            if (subkey)
-                keyids.push_back(subkey->keyid);
-            gpgme_key_release(key);
-        }
-        if (gpg_err_code(gpgErr) != GPG_ERR_EOF)
-        {
-            gpgme_op_keylist_end(ctx);
-            auto msg = tfm::format(_("can not list keys: %s"), gpgme_strerror(gpgErr));
-            logger->debug(msg);
-            throw LrException(LRE_GPGERROR, msg);
-        }
-    }
-    return keyids;
-}
 
 // download key from URL
 std::vector<Key> Repo::Impl::retrieve(const std::string & url)
@@ -821,43 +668,12 @@ std::vector<Key> Repo::Impl::retrieve(const std::string & url)
         throw RepoError(msg);
     }
     lseek(fd, SEEK_SET, 0);
-    auto keyInfos = rawkey2infos(fd);
+    auto keyInfos = Key::keysFromFd(fd);
     for (auto & key : keyInfos)
-        key.url = url;
+        key.setUrl(url);
     return keyInfos;
 }
 
-/*
- * Creates the '/run/user/$UID' directory if it doesn't exist. If this
- * directory exists, gpgagent will create its sockets under
- * '/run/user/$UID/gnupg'.
- *
- * If this directory doesn't exist, gpgagent will create its sockets in gpg
- * home directory, which is under '/var/cache/yum/metadata/' and this was
- * causing trouble with container images, see [1].
- *
- * Previous solution was to send the agent a "KILLAGENT" message, but that
- * would cause a race condition with calling gpgme_release(), see [2], [3],
- * [4].
- *
- * Since the agent doesn't clean up its sockets properly, by creating this
- * directory we make sure they are in a place that is not causing trouble with
- * container images.
- *
- * [1] https://bugzilla.redhat.com/show_bug.cgi?id=1650266
- * [2] https://bugzilla.redhat.com/show_bug.cgi?id=1769831
- * [3] https://github.com/rpm-software-management/microdnf/issues/50
- * [4] https://bugzilla.redhat.com/show_bug.cgi?id=1781601
- */
-static void ensure_socket_dir_exists() {
-    auto logger(Log::getLogger());
-    std::string dirname = "/run/user/" + std::to_string(getuid());
-    int res = mkdir(dirname.c_str(), 0700);
-    if (res != 0 && errno != EEXIST) {
-        logger->debug(tfm::format("Failed to create directory \"%s\": %d - %s",
-                                  dirname, errno, strerror(errno)));
-    }
-}
 
 void Repo::Impl::importRepoKeys()
 {
@@ -865,7 +681,6 @@ void Repo::Impl::importRepoKeys()
 
     auto gpgDir = getCachedir() + "/pubring";
     auto knownKeys = keyidsFromPubring(gpgDir);
-    ensure_socket_dir_exists();
     for (const auto & gpgkeyUrl : conf->gpgkey().getValue()) {
         auto keyInfos = retrieve(gpgkeyUrl);
         for (auto & keyInfo : keyInfos) {
@@ -876,27 +691,21 @@ void Repo::Impl::importRepoKeys()
 
             if (callbacks) {
                 if (!callbacks->repokeyImport(keyInfo.getId(), keyInfo.getUserId(), keyInfo.getFingerprint(),
-                                              keyInfo.url, keyInfo.getTimestamp()))
+                                              keyInfo.getUrl(), keyInfo.getTimestamp()))
                     continue;
             }
 
             struct stat sb;
-            if (stat(gpgDir.c_str(), &sb) != 0 || !S_ISDIR(sb.st_mode))
-                mkdir(gpgDir.c_str(), 0777);
-
-            gpgme_ctx_t ctx;
-            gpgme_new(&ctx);
-            std::unique_ptr<std::remove_pointer<gpgme_ctx_t>::type> context(ctx);
-
-            // set GPG home dir
-            auto gpgErr = gpgme_ctx_set_engine_info(ctx, GPGME_PROTOCOL_OpenPGP, NULL, gpgDir.c_str());
-            if (gpgErr != GPG_ERR_NO_ERROR) {
-                auto msg = tfm::format(_("%s: gpgme_ctx_set_engine_info(): %s"), __func__, gpgme_strerror(gpgErr));
-                logger->debug(msg);
-                throw LrException(LRE_GPGERROR, msg);
+            if (stat(gpgDir.c_str(), &sb) != 0 || !S_ISDIR(sb.st_mode)) {
+                int res = mkdir(gpgDir.c_str(), 0777);
+                if (res != 0 && errno != EEXIST) {
+                    auto msg = tfm::format(_("Failed to create directory \"%s\": %d - %s"),
+                                           gpgDir, errno, strerror(errno));
+                    throw RepoError(msg);
+                }
             }
 
-            gpgImportKey(ctx, keyInfo.rawKey);
+            importKeyToPubring(keyInfo.getAsciiArmoredKey(), gpgDir);
 
             logger->debug(tfm::format(_("repo %s: imported key 0x%s."), id, keyInfo.getId()));
         }
@@ -963,11 +772,15 @@ std::unique_ptr<LrResult> Repo::Impl::lrHandlePerform(LrHandle * handle, const s
     return result;
 }
 
-bool Repo::Impl::loadCache(bool throwExcept)
+bool Repo::Impl::loadCache(bool throwExcept, bool ignoreMissing)
 {
     std::unique_ptr<LrHandle> h(lrHandleInitLocal());
     std::unique_ptr<LrResult> r;
 
+    if (ignoreMissing) {
+        handleSetOpt(h.get(), LRO_IGNOREMISSING, 1L);
+    }
+
     // Fetch data
     try {
         r = lrHandlePerform(h.get(), getCachedir(), conf->repo_gpgcheck().getValue());
@@ -1004,8 +817,13 @@ bool Repo::Impl::loadCache(bool throwExcept)
     for (auto elem = yum_repomd->distro_tags; elem; elem = g_slist_next(elem)) {
         if (elem->data) {
             auto distroTag = static_cast<LrYumDistroTag *>(elem->data);
-            if (distroTag->tag)
-                distro_tags.emplace_back(distroTag->cpeid, distroTag->tag);
+            if (distroTag->tag) {
+                std::string cpeid_str;
+                if (distroTag->cpeid) {
+                    cpeid_str = distroTag->cpeid;
+                }
+                distro_tags.emplace_back(std::move(cpeid_str), distroTag->tag);
+            }
         }
     }
 
@@ -1047,6 +865,9 @@ void Repo::Impl::addCountmeFlag(LrHandle *handle) {
      * This is to align the time window with an absolute point in time rather
      * than the last counting event (which could facilitate tracking across
      * multiple such events).
+     *
+     * In the below comments, the window's current position will be referred to
+     * as "this window" for brevity.
      */
     auto logger(Log::getLogger());
 
@@ -1071,7 +892,7 @@ void Repo::Impl::addCountmeFlag(LrHandle *handle) {
     // Load the cookie
     std::string fname = getPersistdir() + "/" + COUNTME_COOKIE;
     int ver = COUNTME_VERSION;      // file format version (for future use)
-    time_t epoch = 0;               // position of first-ever counted window
+    time_t epoch = 0;               // position of first observed window
     time_t win = COUNTME_OFFSET;    // position of last counted window
     int budget = -1;                // budget for this window (-1 = generate)
     std::ifstream(fname) >> ver >> epoch >> win >> budget;
@@ -1097,8 +918,15 @@ void Repo::Impl::addCountmeFlag(LrHandle *handle) {
 
         // Compute the position of this window
         win = now - (delta % COUNTME_WINDOW);
+
+        // Compute the epoch from this system's epoch or, if unknown, declare
+        // this window as the epoch (unless stored in the cookie previously).
+        time_t sysepoch = getSystemEpoch();
+        if (sysepoch)
+            epoch = sysepoch - ((sysepoch - COUNTME_OFFSET) % COUNTME_WINDOW);
         if (!epoch)
             epoch = win;
+
         // Window step (0 at epoch)
         int step = (win - epoch) / COUNTME_WINDOW;
 
@@ -1107,7 +935,7 @@ void Repo::Impl::addCountmeFlag(LrHandle *handle) {
         for (i = 0; i < COUNTME_BUCKETS.size(); ++i)
             if (step < COUNTME_BUCKETS[i])
                 break;
-        int bucket = i + 1;  // Buckets are indexed from 1
+        int bucket = i + 1;  // Buckets are numbered from 1
 
         // Set the flag
         std::string flag = "countme=" + std::to_string(bucket);
@@ -1130,7 +958,11 @@ bool Repo::Impl::isMetalinkInSync()
 {
     auto logger(Log::getLogger());
     char tmpdir[] = "/tmp/tmpdir.XXXXXX";
-    mkdtemp(tmpdir);
+    if (!mkdtemp(tmpdir)) {
+        const char * errTxt = strerror(errno);
+        throw RepoError(tfm::format(_("Cannot create repo temporary directory \"%s\": %s"),
+                                      tmpdir, errTxt));
+    }
     Finalizer tmpDirRemover([&tmpdir](){
         dnf_remove_recursive(tmpdir, NULL);
     });
@@ -1200,7 +1032,11 @@ bool Repo::Impl::isRepomdInSync()
     auto logger(Log::getLogger());
     LrYumRepo *yum_repo;
     char tmpdir[] = "/tmp/tmpdir.XXXXXX";
-    mkdtemp(tmpdir);
+    if (!mkdtemp(tmpdir)) {
+        const char * errTxt = strerror(errno);
+        throw RepoError(tfm::format(_("Cannot create repo temporary directory \"%s\": %s"),
+                                      tmpdir, errTxt));
+    }
     Finalizer tmpDirRemover([&tmpdir](){
         dnf_remove_recursive(tmpdir, NULL);
     });
@@ -1322,6 +1158,7 @@ bool Repo::Impl::load()
         fetch(cacheDir, lrHandleInitRemote(nullptr));
         timestamp = -1;
         loadCache(true);
+        fresh = true;
     } catch (const LrExceptionWithSourceUrl & e) {
         auto msg = tfm::format(_("Failed to download metadata for repo '%s': %s"), id, e.what());
         throw RepoError(msg);
@@ -1371,7 +1208,7 @@ std::string Repo::Impl::getCachedir() const
 
 std::string Repo::Impl::getPersistdir() const
 {
-    auto persdir(conf->getMasterConfig().persistdir().getValue());
+    auto persdir(conf->getMainConfig().persistdir().getValue());
     if (persdir.back() != '/')
         persdir.push_back('/');
     std::string result = persdir + "repos/" + getHash();
@@ -1383,6 +1220,31 @@ std::string Repo::Impl::getPersistdir() const
     return result;
 }
 
+/* Returns this system's installation time ("epoch") as a UNIX timestamp.
+ *
+ * Uses the machine-id(5) file's mtime as a good-enough source of truth.  This
+ * file is typically tied to the system's installation or first boot where it's
+ * populated by an installer tool or init system, respectively, and is never
+ * changed afterwards.
+ *
+ * Some systems, such as containers that don't run an init system, may have the
+ * file missing, empty or uninitialized, in which case this function returns 0.
+ */
+time_t Repo::Impl::getSystemEpoch() const
+{
+    std::string filename = "/etc/machine-id";
+    std::string id;
+    struct stat st;
+
+    if (stat(filename.c_str(), &st) != 0 || !st.st_size)
+        return 0;
+    std::ifstream(filename) >> id;
+    if (id == "uninitialized")
+        return 0;
+
+    return st.st_mtime;
+}
+
 int Repo::Impl::getAge() const
 {
     return time(NULL) - mtime(getMetadataPath(MD_TYPE_PRIMARY).c_str());
@@ -1451,14 +1313,14 @@ const char * const * Repo::Impl::getHttpHeaders() const
 
 bool Repo::fresh()
 {
-    return pImpl->timestamp >= 0;
+    return pImpl->fresh;
 }
 
 void Repo::Impl::resetMetadataExpired()
 {
     if (expired || conf->metadata_expire().getValue() == -1)
         return;
-    if (conf->getMasterConfig().check_config_file_age().getValue() &&
+    if (conf->getMainConfig().check_config_file_age().getValue() &&
         !repoFilePath.empty() &&
         mtime(repoFilePath.c_str()) > mtime(getMetadataPath(MD_TYPE_PRIMARY).c_str()))
         expired = true;
@@ -1515,6 +1377,18 @@ void Repo::Impl::detachLibsolvRepo()
         attachLibsolvMutex.unlock();
 }
 
+LrAuth Repo::Impl::stringToProxyAuthMethods(const std::string & proxyAuthMethodStr) noexcept
+{
+    auto proxyAuthMethods = LR_AUTH_ANY;
+    for (auto & auth : PROXYAUTHMETHODS) {
+        if (proxyAuthMethodStr == auth.name) {
+            proxyAuthMethods = auth.code;
+            break;
+        }
+    }
+    return proxyAuthMethods;
+}
+
 void Repo::setMaxMirrorTries(int maxMirrorTries)
 {
     pImpl->maxMirrorTries = maxMirrorTries;
@@ -1679,45 +1553,7 @@ static LrHandle * newHandle(ConfigMain * conf)
     // see dnf.repo.Repo._handle_new_remote() how to pass
     if (conf) {
         user_agent = conf->user_agent().getValue().c_str();
-        auto minrate = conf->minrate().getValue();
-        handleSetOpt(h, LRO_LOWSPEEDLIMIT, static_cast<long>(minrate));
-
-        auto maxspeed = conf->throttle().getValue();
-        if (maxspeed > 0 && maxspeed <= 1)
-            maxspeed *= conf->bandwidth().getValue();
-        if (maxspeed != 0 && maxspeed < minrate)
-            throw RepoError(_("Maximum download speed is lower than minimum. "
-                              "Please change configuration of minrate or throttle"));
-        handleSetOpt(h, LRO_MAXSPEED, static_cast<int64_t>(maxspeed));
-
-        if (!conf->proxy().empty() && !conf->proxy().getValue().empty())
-            handleSetOpt(h, LRO_PROXY, conf->proxy().getValue().c_str());
-
-        //set proxy authorization method
-        auto proxyAuthMethodStr = conf->proxy_auth_method().getValue();
-        auto proxyAuthMethod = LR_AUTH_ANY;
-        for (auto & auth : PROXYAUTHMETHODS) {
-            if (proxyAuthMethodStr == auth.name) {
-                proxyAuthMethod = auth.code;
-                break;
-            }
-        }
-        handleSetOpt(h, LRO_PROXYAUTHMETHODS, static_cast<long>(proxyAuthMethod));
-
-        if (!conf->proxy_username().empty()) {
-            auto userpwd = conf->proxy_username().getValue();
-            if (!userpwd.empty()) {
-                if (conf->proxy_password().empty()) {
-                    throw RepoError(_("'proxy_username' is set but not 'proxy_password'"));
-                }
-                userpwd = formatUserPassString(userpwd, conf->proxy_password().getValue(), true);
-                handleSetOpt(h, LRO_PROXYUSERPWD, userpwd.c_str());
-            }
-        }
-
-        auto sslverify = conf->sslverify().getValue() ? 1L : 0L;
-        handleSetOpt(h, LRO_SSLVERIFYHOST, sslverify);
-        handleSetOpt(h, LRO_SSLVERIFYPEER, sslverify);
+        setHandle(h, *conf);
     }
     handleSetOpt(h, LRO_USERAGENT, user_agent);
     return h;
