@@ -375,6 +375,33 @@ add_protected(_GoalObject *self, PyObject *seq) try
     Py_RETURN_NONE;
 } CATCH_TO_PYTHON
 
+static PyObject *
+add_exclude_from_weak(_GoalObject *self, PyObject *seq) try
+{
+    HyGoal goal = self->goal;
+    auto pset = pyseq_to_packageset(seq, hy_goal_get_sack(goal));
+    if (!pset)
+        return NULL;
+    goal->add_exclude_from_weak(*(pset.get()));
+    Py_RETURN_NONE;
+} CATCH_TO_PYTHON
+
+static PyObject *
+reset_exclude_from_weak(_GoalObject *self, PyObject *unused) try
+{
+    HyGoal goal = self->goal;
+    goal->reset_exclude_from_weak();
+    Py_RETURN_NONE;
+} CATCH_TO_PYTHON
+
+static PyObject *
+exclude_from_weak_autodetect(_GoalObject *self, PyObject *unused) try
+{
+    HyGoal goal = self->goal;
+    goal->exclude_from_weak_autodetect();
+    Py_RETURN_NONE;
+} CATCH_TO_PYTHON
+
 static PyObject *
 run(_GoalObject *self, PyObject *args, PyObject *kwds) try
 {
@@ -451,6 +478,12 @@ problem_broken_dependency(_GoalObject *self, PyObject *args, PyObject *kwds) try
     return packageset_to_pylist(pset.get(), self->sack);
 } CATCH_TO_PYTHON
 
+static PyObject *
+file_dep_problem_present(_GoalObject *self, PyObject *unused) try
+{
+    return PyBool_FromLong(self->goal->isBrokenFileDependencyPresent());
+} CATCH_TO_PYTHON
+
 static PyObject *
 log_decisions(_GoalObject *self, PyObject *unused) try
 {
@@ -597,6 +630,9 @@ static struct PyMethodDef goal_methods[] = {
      NULL},
     {"add_protected", (PyCFunction)add_protected, METH_O,
      NULL},
+    {"add_exclude_from_weak", (PyCFunction)add_exclude_from_weak, METH_O, NULL},
+    {"reset_exclude_from_weak", (PyCFunction)reset_exclude_from_weak, METH_NOARGS, NULL},
+    {"exclude_from_weak_autodetect", (PyCFunction)exclude_from_weak_autodetect, METH_NOARGS, NULL},
     {"distupgrade_all",        (PyCFunction)distupgrade_all,        METH_NOARGS,        NULL},
     {"distupgrade",                (PyCFunction)distupgrade,
      METH_VARARGS | METH_KEYWORDS, NULL},
@@ -622,6 +658,7 @@ static struct PyMethodDef goal_methods[] = {
     {"problem_conflicts",(PyCFunction)problem_conflicts,        METH_VARARGS | METH_KEYWORDS,                NULL},
     {"problem_broken_dependency",(PyCFunction)problem_broken_dependency,        METH_VARARGS | METH_KEYWORDS,                NULL},
     {"problem_rules", (PyCFunction)problem_rules,        METH_NOARGS,                NULL},
+    {"file_dep_problem_present", (PyCFunction)file_dep_problem_present,        METH_NOARGS,        NULL},
     {"log_decisions",   (PyCFunction)log_decisions,        METH_NOARGS,        NULL},
     {"write_debugdata", (PyCFunction)write_debugdata,        METH_O,                NULL},
     {"list_erasures",        (PyCFunction)list_erasures,        METH_NOARGS,        NULL},
