@@ -18,11 +18,6 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
-#include <check.h>
-#include <glib.h>
-#include <stdarg.h>
-#include <vector>
-
 #include "libdnf/goal/Goal.hpp"
 #include "libdnf/dnf-types.h"
 #include "libdnf/hy-goal-private.hpp"
@@ -37,10 +32,16 @@
 #include "libdnf/hy-selector.h"
 #include "libdnf/hy-util-private.hpp"
 #include "libdnf/sack/packageset.hpp"
+
 #include "fixtures.h"
 #include "testsys.h"
 #include "test_suites.h"
 
+#include <check.h>
+#include <glib.h>
+#include <stdarg.h>
+#include <vector>
+
 static DnfPackage *
 get_latest_pkg(DnfSack *sack, const char *name)
 {
@@ -49,8 +50,8 @@ get_latest_pkg(DnfSack *sack, const char *name)
     hy_query_filter(q, HY_PKG_REPONAME, HY_NEQ, HY_SYSTEM_REPO_NAME);
     hy_query_filter_latest_per_arch(q, 1);
     GPtrArray *plist = hy_query_run(q);
-    fail_unless(plist->len == 1,
-                "get_latest_pkg() failed finding '%s'.", name);
+    ck_assert_msg(plist->len == 1,
+                  "get_latest_pkg() failed finding '%s'.", name);
     auto pkg = static_cast<DnfPackage *>(g_object_ref(g_ptr_array_index(plist, 0)));
     hy_query_free(q);
     g_ptr_array_unref(plist);
@@ -435,7 +436,7 @@ assert_list_names(bool wanted, GPtrArray *plist, ...)
     va_start(names, plist);
     while ((name = va_arg(names, char *)) != NULL) {
         if (i++ >= count) {
-            fail("assert_list_names(): list too short");
+            ck_abort_msg("assert_list_names(): list too short");
         }
         bool found = false;
         for (auto string: stringVector) {
@@ -445,8 +446,8 @@ assert_list_names(bool wanted, GPtrArray *plist, ...)
             }
         }
         if ((wanted && !found) || (!wanted && found)) {
-            fail_unless(false, "assert_list_names(): element '%s' %sfound '%zu'",
-                        name, wanted ? "not ": "", stringVector.size());
+            ck_abort_msg("assert_list_names(): element '%s' %sfound '%zu'",
+                         name, wanted ? "not ": "", stringVector.size());
         }
     }
     // In the wanted case; we expect all the pkgs in the lists to fully
@@ -454,7 +455,7 @@ assert_list_names(bool wanted, GPtrArray *plist, ...)
     // all the passed pkg arguments are *not* found in the list, which is
     // already checked above.
     if (wanted) {
-        fail_unless(i == count, "assert_list_names(): too many items in the list (%d vs %d)", i, count);
+        ck_assert_msg(i == count, "assert_list_names(): too many items in the list (%d vs %d)", i, count);
     }
     va_end(names);
 }
@@ -492,7 +493,7 @@ START_TEST(test_goal_upgrade_all)
     g_ptr_array_unref(plist_obs);
     g_ptr_array_unref(plist);
 
-    fail_unless(size_and_free(hy_goal_list_installs(goal, NULL)) == 0);
+    ck_assert(size_and_free(hy_goal_list_installs(goal, NULL)) == 0);
     hy_goal_free(goal);
 }
 END_TEST
@@ -593,11 +594,11 @@ START_TEST(test_goal_describe_problem_rules)
     auto problems = goal->describeProblemRules(0, true);
     const char *expected[] = {
                 "conflicting requests",
-                "nothing provides goodbye needed by hello-1-1.noarch"
+                "nothing provides goodbye needed by hello-1-1.noarch from main"
                 };
-    fail_unless(problems.size() == 2);
-    fail_unless(problems[0] == expected[0]);
-    fail_unless(problems[1] == expected[1]);
+    ck_assert_int_eq(problems.size(), 2);
+    ck_assert_str_eq(problems[0].c_str(), expected[0]);
+    ck_assert_str_eq(problems[1].c_str(), expected[1]);
 
     g_object_unref(pkg);
     hy_goal_free(goal);
@@ -752,7 +753,7 @@ START_TEST(test_goal_favor)
     HySelector sltr = hy_selector_create(sack);
 
     hy_selector_set(sltr, HY_PKG_NAME, HY_EQ, "dodo");
-    hy_goal_install_selector(goal, sltr, NULL);
+    fail_if(!hy_goal_install_selector(goal, sltr, NULL));
     DnfPackage *pkg = get_latest_pkg(sack, "dodo-dep-a");
     hy_goal_favor(goal, pkg);
     fail_if(hy_goal_run_flags(goal, DNF_NONE));
@@ -774,7 +775,7 @@ START_TEST(test_goal_disfavor)
     HySelector sltr = hy_selector_create(sack);
 
     hy_selector_set(sltr, HY_PKG_NAME, HY_EQ, "dodo");
-    hy_goal_install_selector(goal, sltr, NULL);
+    fail_if(!hy_goal_install_selector(goal, sltr, NULL));
     DnfPackage *pkg = get_latest_pkg(sack, "dodo-dep-a");
     hy_goal_disfavor(goal, pkg);
     fail_if(hy_goal_run_flags(goal, DNF_NONE));
@@ -860,10 +861,10 @@ START_TEST(test_goal_lock)
 
     auto problems = goal->describeProblemRules(0, true);
     const char *expected[] = {
-                "package bloop-ext-2.0-1.noarch requires bloop = 2.0-1, but none of the providers can be installed",
-                "cannot install both bloop-2.0-1.noarch and bloop-1.0-1.noarch",
+                "package bloop-ext-2.0-1.noarch from updates requires bloop = 2.0-1, but none of the providers can be installed",
+                "cannot install both bloop-2.0-1.noarch from updates and bloop-1.0-1.noarch from @System",
                 "conflicting requests",
-                "package bloop-ext-1.0-1.noarch is filtered out by exclude filtering"
+                "package bloop-ext-1.0-1.noarch from updates is filtered out by exclude filtering"
                 };
     ck_assert_int_eq(problems.size(), 4);
     ck_assert_str_eq(problems[0].c_str(), expected[0]);
@@ -986,7 +987,7 @@ START_TEST(test_goal_describe_problem_excludes)
     HySelector sltr = hy_selector_create(sack);
 
     hy_selector_set(sltr, HY_PKG_NAME, HY_EQ, "semolina");
-    hy_goal_install_selector(goal, sltr, NULL);
+    fail_if(!hy_goal_install_selector(goal, sltr, NULL));
     hy_selector_free(sltr);
 
     fail_unless(hy_goal_run_flags(goal, DNF_NONE));
